'use strict';

var chai = require('chai');
var sinon = require('sinon');
var sinonChai = require('sinon-chai');
var expect = chai.expect;
chai.use(sinonChai);

var energyQuery = require('../../lib/energy-query');
var EnergyTable = require('../../lib/energy-table').EnergyTable;
var mocks = require('../fixtures/table-mocks');

describe('EnergyQuery (class)', function() {

  var EnergyQueryFactory = energyQuery.EnergyQueryFactory;

  var table;
  var tableHashAndRange;

  before(function(done) {
    table = new EnergyTable(mocks.dbMock, 'Table-HashKey');
    table.init(done);
  });

  before(function(done) {
    tableHashAndRange = new EnergyTable(mocks.dbMock, 'Table-HashKey-RangeKey');
    tableHashAndRange.init(done);
  });

  describe('#getQuery()', function() {

    it('should return the correct query for "insert" operations', function(done) {
      var item = {
        'key-0': 'value-0',
        'key-1': 'value-1',
        'key-2': 12345
      };

      var expectedQuery = {
        TableName: 'Table-HashKey',
        Item: {
          'key-0': { S: 'value-0' },
          'key-1': { S: 'value-1' },
          'key-2': { N: '12345' }
        }
      };

      var instance = new EnergyQueryFactory(table);

      instance.getInsertQuery(item, function(err, query) {
        if (err) return done(err);
        expect(query).to.deep.equals(expectedQuery);
        done();
      });

    });

    it('should return the correct query for "query" operations', function(done) {
      var doc = {
        'key-0': 'value-0',
        'key-1': {'$gte': 12345},
        'key-2': {
          'sub-key': 'random-value'
        }
      };

      var expectedQuery = {
        TableName: 'Table-HashKey',
        ExpressionAttributeNames: {
          '#k0': 'key-0',
          '#k1': 'key-1',
          '#k2': 'key-2'
        },
        ExpressionAttributeValues: {
          ':v0': {S: 'value-0'},
          ':v1': {N: '12345'},
          ':v2': {M: {
            'sub-key': {S: 'random-value'}
          }}
        },
        KeyConditionExpression: '#k0 = :v0 AND #k1 >= :v1 AND #k2 = :v2'
      };

      var instance = new EnergyQueryFactory(table);

      instance.getQuery(doc, function(err, query) {
        if (err) return done(err);
        expect(query).to.deep.equals(expectedQuery);
        done();
      });

    });

    it('should return the correct query for "scan" operations', function(done) {
      var doc = {
        'key-0': 'value-0',
        'key-1': {'$gte': 12345},
        'key-2': {
          'sub-key': 'random-value'
        }
      };

      var expectedQuery = {
        TableName: 'Table-HashKey',
        ExpressionAttributeNames: {
          '#k0': 'key-0',
          '#k1': 'key-1',
          '#k2': 'key-2'
        },
        ExpressionAttributeValues: {
          ':v0': {S: 'value-0'},
          ':v1': {N: '12345'},
          ':v2': {M: {
            'sub-key': {S: 'random-value'}
          }}
        },
        FilterExpression: '#k0 = :v0 AND #k1 >= :v1 AND #k2 = :v2'
      };

      var instance = new EnergyQueryFactory(table);

      instance.getScanQuery(doc, function(err, query) {
        if (err) return done(err);
        expect(query).to.deep.equals(expectedQuery);
        done();
      });

    });

    it('should return the correct query for full "scan" operations', function(done) {
      var doc = {};

      var expectedQuery = {
        TableName: 'Table-HashKey',
      };

      var instance = new EnergyQueryFactory(table);

      instance.getScanQuery(doc, function(err, query) {
        if (err) return done(err);
        expect(query).to.deep.equals(expectedQuery);
        done();
      });

    });

    it('should return the correct query for "delete" operations', function(done) {
      var doc = {
        'some-hash-key': 'value-0',
        'some-range-key': 'value-1',
        'key-2': {'$gte': 12345},
        'key-3': {
          'sub-key': 'random-value'
        }
      };

      var expectedQuery = {
        TableName: 'Table-HashKey-RangeKey',
        Key: {
          'some-hash-key': {S: 'value-0'},
          'some-range-key': {S: 'value-1'},
        },
        ExpressionAttributeNames: {
          '#k0': 'some-hash-key',
          '#k1': 'some-range-key',
          '#k2': 'key-2',
          '#k3': 'key-3'
        },
        ExpressionAttributeValues: {
          ':v0': {S: 'value-0'},
          ':v1': {S: 'value-1'},
          ':v2': {N: '12345'},
          ':v3': {M: {
            'sub-key': {S: 'random-value'}
          }}
        },
        ConditionExpression:
          '#k0 = :v0 AND #k1 = :v1 AND #k2 >= :v2 AND #k3 = :v3'
      };

      var instance = new EnergyQueryFactory(tableHashAndRange);

      instance.getDeleteQuery(doc, function(err, query) {
        if (err) return done(err);
        expect(query).to.deep.equals(expectedQuery);
        done();
      });

    });

    it('should return the correct query for "update" operations', function(done) {
      var doc = {
        'some-hash-key': 'value-0',
        'some-range-key': 'value-1',
        'key-0': 'value-2',
        'key-1': 98765,
        'key-2': 12345,
      };

      var update = {
        '$set': {
          'key-0': 'value-0-new',
          'key-1': 11111,
        },
        '$inc': {
          'key-2': 3
        },
        '$unset': {
          'key-3': 1
        },
      };

      var expectedQuery = {
        TableName: 'Table-HashKey-RangeKey',
        Key: {
          'some-hash-key': {S: 'value-0'},
          'some-range-key': {S: 'value-1'},
        },
        ExpressionAttributeNames: {
          '#k0': 'some-hash-key',
          '#k1': 'some-range-key',
          '#k2': 'key-0',
          '#k3': 'key-1',
          '#k4': 'key-2',
          '#k5': 'key-0',
          '#k6': 'key-1',
          '#k7': 'key-2',
          '#k8': 'key-3',
        },
        ExpressionAttributeValues: {
          ':v0': {S: 'value-0'},
          ':v1': {S: 'value-1'},
          ':v2': {S: 'value-2'},
          ':v3': {N: '98765'},
          ':v4': {N: '12345'},
          ':v5': {S: 'value-0-new'},
          ':v6': {N: '11111'},
          ':v7': {N: '3'},
        },
        ConditionExpression:
          '#k0 = :v0 AND #k1 = :v1 AND #k2 = :v2 AND #k3 = :v3 AND #k4 = :v4',
        UpdateExpression:
          'SET #k5 = :v5, #k6 = :v6 ADD #k7 :v7 REMOVE #k8'
      };

      var instance = new EnergyQueryFactory(tableHashAndRange);

      instance.getUpdateQuery(doc, update, function(err, query) {
        if (err) return done(err);
        expect(query).to.deep.equals(expectedQuery);
        done();
      });

    });

    it('should return the correct query for "replace" operations', function(done) {
      var doc = {
        'key-0': 'value-0',
        'key-1': 98765,
        'key-2': {'$gt': 12345},
      };

      var replacement = {
        'key-0': 'value-0-new',
        'key-1': 55555,
        'key-2': 99999,
      };

      var expectedQuery = {
        TableName: 'Table-HashKey',
        Item: {
          'key-0': {S: 'value-0-new'},
          'key-1': {N: '55555'},
          'key-2': {N: '99999'},
        },
        ExpressionAttributeNames: {
          '#k0': 'key-0',
          '#k1': 'key-1',
          '#k2': 'key-2',
        },
        ExpressionAttributeValues: {
          ':v0': {S: 'value-0'},
          ':v1': {N: '98765'},
          ':v2': {N: '12345'},
        },
        ConditionExpression:
          '#k0 = :v0 AND #k1 = :v1 AND #k2 > :v2',
      };

      var instance = new EnergyQueryFactory(table);

      instance.getReplaceQuery(doc, replacement, function(err, query) {
        if (err) return done(err);
        expect(query).to.deep.equals(expectedQuery);
        done();
      });

    });
  });
});

describe('EnergyQuery (module)', function() {

  describe('#getExpressionFragment(expressionName, expressionValue, value)',
    function() {

      it('should convert to the respective expression fragment', function() {

        var f = function (test) {
          return energyQuery.getExpressionFragment('#k0', ':v0', test);
        }

        expect(f('some string')).to.equals('#k0 = :v0');
        expect(f(12345)).to.equals('#k0 = :v0');
        expect(f({key: 'random object'})).to.equals('#k0 = :v0');

        expect(f({$eq: 12345})).to.equals('#k0 = :v0');
        expect(f({$gt: 12345})).to.equals('#k0 > :v0');
        expect(f({$gte: 12345})).to.equals('#k0 >= :v0');
        expect(f({$lt: 12345})).to.equals('#k0 < :v0');
        expect(f({$lte: 12345})).to.equals('#k0 <= :v0');

      });

    }
  );

  describe('#getExpressionValue(value)',
    function() {

      it('should convert to the respective expression fragment', function() {

        var f = energyQuery.getExpressionValue;

        expect(f('some string')).to.deep.equals({S: 'some string'});
        expect(f(12345)).to.deep.equals({N: '12345'});
        expect(f({key: 'random object'})).to.deep.equals(
          {M: {key: {S: 'random object'}}}
        );

        expect(f({$eq: 12345})).to.deep.equals({N: '12345'});
        expect(f({$gt: 12345})).to.deep.equals({N: '12345'});
        expect(f({$gte: 12345})).to.deep.equals({N: '12345'});
        expect(f({$lt: 12345})).to.deep.equals({N: '12345'});
        expect(f({$lte: 12345})).to.deep.equals({N: '12345'});

      });

    }
  );

});
